import fs from 'fs';
import path from 'path';
import chalk from 'chalk';
import * as yargs from 'yargs';
import remark from 'remark';
import toc from 'remark-toc';
import { isHeadingNode, isParagraphNode, isBlockNode, isCollapsibleNode, isBoldNode, isEmphasisNode, isLinkNode, isInlineCodeNode, isCodeNode, isTextNode, isTableCellNode, isTableRowNode, isTableNode, isNodeWithChildren, isNodeWithValue, propsToDocumentation, apiDocsConfig, mergeConfig } from '@structured-types/api-docs';
import reactPlugin from '@structured-types/react-plugin';
import propTypesPlugin from '@structured-types/prop-types-plugin';
import { parseFiles } from '@structured-types/api';

const trimQuotes = (txt) => txt ? txt.replace(/['"]+/g, '') : txt;
const extractCustomTag = (node, tagName) => {
    const nodes = node.children &&
        node.children.filter((child) => child.type === 'html' &&
            child.value &&
            child.value.startsWith(`<${tagName}`));
    return nodes
        ? nodes
            .map((section) => ({
            section,
            match: section.value
                ? section.value.match(/(\w+)\s*=\s*((["'])(.*?)\3|([^>\s]*)(?=\s|\/>))(?=[^<]*>)/g)
                : undefined,
        }))
            .map(({ section, match, }) => ({
            attrs: { section, tagName, node },
            attributes: match
                ? match.map((m) => m.split('=').map((value) => trimQuotes(value)))
                : undefined,
        }))
        : undefined;
};
const inlineNewContent = ({ section, tagName, node }, newContent) => {
    const startTag = `<!-- START-${tagName.toUpperCase()} -->`;
    const endTag = `<!-- END-${tagName.toUpperCase()} -->`;
    const index = node.children ? node.children.indexOf(section) : -1;
    if (index === -1) {
        return;
    }
    let deleteNodes = 0;
    let foundComment = false;
    if (node.children) {
        for (let i = index + 1; i < node.children.length; i += 1) {
            const childNode = node.children[i];
            if (childNode.type == 'html' && childNode.value == startTag) {
                foundComment = true;
            }
            if (foundComment) {
                deleteNodes += 1;
            }
            if (childNode &&
                childNode.type == 'html' &&
                (childNode.value == endTag ||
                    (childNode.value &&
                        childNode.value.toLowerCase().startsWith(`<${tagName}`)))) {
                break;
            }
        }
        node.children.splice(index + 1, deleteNodes, ...[
            {
                type: 'html',
                value: startTag,
            },
            ...newContent,
            {
                type: 'html',
                value: endTag,
            },
        ]);
    }
};

const insertOverview = () => {
    return (node) => {
        const sections = extractCustomTag(node, 'package-section');
        if (sections) {
            sections.forEach(({ attrs, attributes }) => {
                if (attributes) {
                    const file = attributes.find((attribute) => attribute[0] === 'file');
                    const sectionAttr = attributes.find((attribute) => attribute[0] === 'section');
                    const sectionName = sectionAttr
                        ? sectionAttr[1].toLowerCase()
                        : 'overview';
                    if (file) {
                        const content = fs.readFileSync(file[1], 'utf8');
                        const overviewNodes = [];
                        const packageJSONFilename = path.resolve(path.join(path.dirname(file[1]), 'package.json'));
                        if (fs.existsSync(packageJSONFilename)) {
                            const packageJSON = JSON.parse(fs.readFileSync(packageJSONFilename, 'utf8'));
                            const { name, description, repository: { directory }, } = packageJSON;
                            const rootRepoDir = `https://github.com/ccontrols/component-controls/blob/master/${directory}`;
                            overviewNodes.push({
                                type: 'heading',
                                depth: 2,
                                children: [
                                    {
                                        type: 'link',
                                        url: rootRepoDir,
                                        children: [{ type: 'text', value: name }],
                                    },
                                ],
                            });
                            overviewNodes.push({
                                type: 'paragraph',
                                children: [{ type: 'text', value: description }],
                            });
                        }
                        remark()
                            .use(() => (sectionNode) => {
                            let foundDepth = 0;
                            if (sectionNode.children) {
                                for (const section of sectionNode.children) {
                                    if (section.type === 'heading') {
                                        if (section.children &&
                                            section.children.length > 0 &&
                                            section.children[0].type === 'text' &&
                                            section.children[0].value &&
                                            section.children[0].value.toLowerCase() === sectionName) {
                                            foundDepth = section.depth;
                                        }
                                        else if (foundDepth >= section.depth) {
                                            break;
                                        }
                                        else if (foundDepth > 0) {
                                            overviewNodes.push(section);
                                        }
                                    }
                                    else if (foundDepth > 0) {
                                        overviewNodes.push(section);
                                    }
                                }
                            }
                        })
                            .process(content);
                        inlineNewContent(attrs, overviewNodes);
                    }
                }
            });
        }
    };
};

const renderNode = ({ node, ...props }, skipLineBreaks = false) => {
    if (isHeadingNode(node)) {
        return nodeContent({ node, as: 'heading', depth: node.depth, ...props });
    }
    else if (isParagraphNode(node)) {
        return nodeContent({ node, as: 'paragraph', ...props });
    }
    else if (isBlockNode(node) && node.children && !skipLineBreaks) {
        return {
            type: 'paragraph',
            children: [
                ...node.children.map((node) => renderNode({ node })),
                {
                    type: 'html',
                    value: '<br />',
                },
            ],
        };
    }
    else if (isCollapsibleNode(node)) {
        return {
            type: 'paragraph',
            children: [
                {
                    type: 'html',
                    value: '<details>',
                },
                {
                    type: 'html',
                    value: '<summary>',
                },
                ...nodesToRemark(node.summary),
                {
                    type: 'html',
                    value: '</summary>',
                },
                {
                    type: 'html',
                    value: '<blockquote>',
                },
                ...(node.children
                    ? node.children.reduce((acc, n, idx) => {
                        const r = renderNode({ node: n }, idx === node.children.length - 1);
                        if (r) {
                            const returns = [r];
                            return [...acc, ...returns];
                        }
                        return acc;
                    }, [])
                    : []),
                {
                    type: 'html',
                    value: '</blockquote>',
                },
                {
                    type: 'html',
                    value: '</details>',
                },
            ],
        };
    }
    else if (isBoldNode(node)) {
        return nodeContent({ node, as: 'strong', ...props });
    }
    else if (isEmphasisNode(node)) {
        return nodeContent({ node, as: 'emphasis', ...props });
    }
    else if (isLinkNode(node)) {
        return node.url
            ? nodeContent({ node, as: 'link', url: node.url, ...props })
            : nodeContent({ node, as: 'paragraph', ...props });
    }
    else if (isInlineCodeNode(node)) {
        return nodeContent({ node, as: 'inlineCode', ...props });
    }
    else if (isCodeNode(node)) {
        return nodeContent({ node, as: 'code', ...props });
    }
    else if (isTextNode(node)) {
        return nodeContent({ node, as: 'text', ...props });
    }
    else if (isTableCellNode(node)) {
        return nodeContent({ node, as: 'tableCell', ...props });
    }
    else if (isTableRowNode(node)) {
        return nodeContent({ node, as: 'tableRow', ...props });
    }
    else if (isTableNode(node)) {
        return nodeContent({ node, as: 'table', ...props });
    }
    else if (isNodeWithChildren(node)) {
        return nodeContent({ node, as: 'paragraph', ...props });
    }
    return null;
};
const nodeContent = ({ node, as, ...rest }) => {
    const value = isNodeWithChildren(node)
        ? nodeComponents({ nodes: node.children, as, ...rest })
        : isNodeWithValue(node)
            ? { type: as, ...rest, value: node.value }
            : null;
    return value;
};
const nodeComponents = ({ nodes, as, ...rest }) => {
    return nodes
        ? {
            type: as,
            children: nodes
                .map((node) => renderNode({ node }))
                .filter((n) => n),
            ...rest,
        }
        : null;
};
const nodesToRemark = (nodes) => {
    return nodes
        .map((node) => {
        const rendered = renderNode({ node });
        return rendered;
    })
        .filter((n) => n);
};

const extractProps = async (files, config) => {
    const props = parseFiles(files, {
        collectSourceInfo: true,
        collectHelpers: true,
        plugins: [propTypesPlugin, reactPlugin],
        ...config,
    });
    const nodes = await propsToDocumentation(props, config);
    return nodesToRemark(nodes);
};

/* eslint-disable prefer-spread */
const insertAPISection = (fileName, configFileName) => () => {
    const resolve = (file, filePath) => {
        if (file.startsWith('.')) {
            return path.resolve(path.dirname(path.resolve(filePath)), file);
        }
        return require.resolve(file);
    };
    const splitCommaAttribute = (attributes, name) => {
        if (attributes) {
            const attribute = attributes.find((attribute) => attribute[0] === name);
            return attribute
                ? attribute[1]
                    .split(',')
                    .map((f) => f.trim())
                    .filter((f) => f)
                : undefined;
        }
        return undefined;
    };
    return async (node) => {
        var _a;
        const sections = extractCustomTag(node, 'api-readme');
        if (sections) {
            for (const section of sections) {
                const { attrs, attributes = [] } = section;
                if (attributes) {
                    const newNodes = [];
                    let elementId = undefined;
                    const inlineOptions = attributes.reduce((acc, attr) => {
                        let value;
                        const strValue = attr[1];
                        if (strValue === 'true') {
                            value = true;
                        }
                        else if (strValue === 'false') {
                            value = false;
                        }
                        else if (!isNaN(parseFloat(strValue))) {
                            value = parseFloat(strValue);
                        }
                        else if (typeof attr[1] === 'string') {
                            if (attr[0] === 'id') {
                                elementId = attr[1];
                                return acc;
                            }
                            value = splitCommaAttribute(attributes, attr[0]);
                        }
                        return { ...acc, [attr[0]]: value };
                    }, { files: [] });
                    const { config = {}, filepath: configFilePath } = (await apiDocsConfig(fileName, configFileName, {
                        elementId,
                    })) || {};
                    const mergedConfig = mergeConfig(config, inlineOptions);
                    const files = configFilePath && config.files
                        ? config.files.map((f) => resolve(f, configFilePath))
                        : (_a = inlineOptions.files) === null || _a === void 0 ? void 0 : _a.map((f) => resolve(f, fileName));
                    if (!files.length) {
                        const srcFolder = resolve('./src', fileName);
                        if (fs.existsSync(srcFolder)) {
                            const sourceFiles = fs.readdirSync(srcFolder);
                            const indexFile = sourceFiles.find((f) => f.startsWith('index'));
                            if (indexFile) {
                                files.push(resolve(path.resolve(srcFolder, indexFile), fileName));
                            }
                        }
                    }
                    const tsNodes = await extractProps(files, mergedConfig);
                    newNodes.push(...tsNodes);
                    inlineNewContent(attrs, newNodes);
                }
            }
        }
    };
};

/**
 * create markdown pages from README.md type files
 * command-line arguments
 * @param {string} f - input file name to be processed, default is README.md
 * @param {boolean} t - whether to generate a table of contents, default is true
 * @param {boolean} l - whether to output logs, default is true
 */
var index = () => {
    const args = yargs
        .help()
        .alias('help', 'h')
        .option('f', {
        alias: 'file',
        describe: 'Input file name',
        type: 'string',
        default: 'README.md',
    })
        .option('c', {
        alias: 'config',
        describe: 'Configuration file full path',
        type: 'string',
    })
        .option('t', {
        alias: 'toc',
        describe: 'Generate a table of content',
        type: 'boolean',
        default: true,
    })
        .option('l', {
        alias: 'log',
        describe: 'Enabled logs',
        type: 'boolean',
        default: true,
    });
    const options = args.parse(process.argv);
    const fileName = options.f;
    const configFileName = options.c;
    if (options.log) {
        console.log('processing file:', chalk.red(path.resolve(fileName)));
    }
    let r = remark()
        .use(insertAPISection(fileName, configFileName))
        .use(insertOverview);
    if (options.toc) {
        r = r.use(toc, { tight: true });
    }
    r.process(fs.readFileSync(fileName, 'utf8'), function (err, file) {
        if (err) {
            throw err;
        }
        fs.writeFileSync(fileName, String(file));
    });
};

export { index as default };
